@@ -4,41 +4,149 @@
 import sys
 import re
 import getopt
+import keyword
 import json
-try:
-    import yaml
-except ImportError:
-    yaml = None
-try:
-    import jsonlines
-except ImportError:
-    jsonlines = None
+
+my_yaml = None
+my_jsonlines = None
 
 import CloudFlare
+from .dump import dump_commands, dump_commands_from_web
 from . import converters
+from . import examples
 
-def dump_commands(cf):
-    """dump a tree of all the known API commands"""
-    w = cf.api_list()
-    sys.stdout.write('\n'.join(w) + '\n')
-
-def dump_commands_from_web(cf):
-    """dump a tree of all the known API commands - from web"""
-    w = cf.api_from_web()
-    for r in w:
-        if r['deprecated']:
-            if r['deprecated_already']:
-                sys.stdout.write('%-6s %s ; deprecated %s - expired!\n' % (r['action'], r['cmd'], r['deprecated_date']))
-            else:
-                sys.stdout.write('%-6s %s ; deprecated %s\n' % (r['action'], r['cmd'], r['deprecated_date']))
-        else:
-            sys.stdout.write('%-6s %s\n' % (r['action'], r['cmd']))
+def load_and_check_yaml():
+    """ load_and_check_yaml() """
+    from . import myyaml
+    global my_yaml
+    my_yaml = myyaml.myyaml()
+    if not my_yaml.available():
+        sys.exit('cli4: install yaml support')
 
 def strip_multiline(s):
     """ remove leading/trailing tabs/spaces on each line"""
     # This hack is needed in order to use yaml.safe_load() on JSON text - tabs are not allowed
     return '\n'.join([l.strip() for l in s.splitlines()])
 
+def process_params_content_files(method, binary_file, args):
+    """ process_params_content_files() """
+
+    digits_only = re.compile('^-?[0-9]+$')
+    floats_only = re.compile('^-?[0-9.]+$')
+
+    params = None
+    content = None
+    files = None
+    # next grab the params. These are in the form of tag=value or =value or @filename
+    while len(args) > 0 and ('=' in args[0] or args[0][0] == '@'):
+        arg = args.pop(0)
+        if arg[0] == '@':
+            # a file to be uploaded - used in workers/script etc - only via PUT or POST
+            filename = arg[1:]
+            if method not in ['PUT','POST']:
+                sys.exit('cli4: %s - raw file upload only with PUT or POST' % (filename))
+            try:
+                if filename == '-':
+                    if binary_file:
+                        content = sys.stdin.buffer.read()
+                    else:
+                        content = sys.stdin.read()
+                else:
+                    if binary_file:
+                        with open(filename, 'rb') as f:
+                            content = f.read()
+                    else:
+                        with open(filename, 'r') as f:
+                            content = f.read()
+            except IOError:
+                sys.exit('cli4: %s - file open failure' % (filename))
+            continue
+        tag_string, value_string = arg.split('=', 1)
+        if value_string.lower() == 'true':
+            value = True
+        elif value_string.lower() == 'false':
+            value = False
+        elif value_string == '' or value_string.lower() == 'none':
+            value = None
+        elif value_string[0] == '=' and value_string[1:] == '':
+            sys.exit('cli4: %s== - no number value passed' % (tag_string))
+        elif value_string[0] == '=' and digits_only.match(value_string[1:]):
+            value = int(value_string[1:])
+        elif value_string[0] == '=' and floats_only.match(value_string[1:]):
+            value = float(value_string[1:])
+        elif value_string[0] == '=':
+            sys.exit('cli4: %s== - invalid number value passed' % (tag_string))
+        elif value_string[0] in '[{' and value_string[-1] in '}]':
+            # a json structure - used in pagerules
+            try:
+                #value = json.loads(value) - changed to yaml code to remove unicode string issues
+                load_and_check_yaml()
+                # cleanup string before parsing so that yaml.safe.load does not complain about whitespace
+                # >>> found character '\t' that cannot start any token <<<
+                value_string = strip_multiline(value_string)
+                try:
+                    value = my_yaml.safe_load(value_string)
+                except my_yaml.parser.ParserError as e:
+                    raise ValueError
+            except ValueError:
+                sys.exit('cli4: %s="%s" - can\'t parse json value' % (tag_string, value_string))
+        elif value_string[0] == '@':
+            # a file to be uploaded - used in dns_records/import etc - only via PUT or POST
+            filename = value_string[1:]
+            if method not in ['PUT', 'POST']:
+                sys.exit('cli4: %s=%s - file upload only with PUT or POST' % (tag_string, filename))
+            if files is None:
+                files = {}
+            if tag_string in files:
+                sys.exit('cli4: %s=%s - duplicate name' % (tag_string, filename))
+            try:
+                if filename == '-':
+                    files[tag_string] = sys.stdin
+                else:
+                    files[tag_string] = open(filename, 'rb')
+            except IOError:
+                sys.exit('cli4: %s=%s - file open failure' % (tag_string, filename))
+            # no need for param code below
+            continue
+        elif (value_string[0] == '"' and value_string[-1] == '"') or (value_string[0] == '\'' and value_string[-1] == '\''):
+            # remove quotes
+            value = value_string[1:-1]
+        else:
+            value = value_string
+
+        if tag_string == '':
+            # There's no tag; it's just an unnamed list
+            if params is None:
+                params = value
+            else:
+                sys.exit('cli4: %s=%s - param error. Can\'t mix unnamed and named list' %
+                         (tag_string, value_string))
+        else:
+            if params is None:
+                params = {}
+            tag = tag_string
+            try:
+                params[tag] = value
+            except TypeError:
+                sys.exit('cli4: %s=%s - param error. Can\'t mix unnamed and named list' %
+                         (tag_string, value_string))
+
+    if content and params:
+        sys.exit('cli4: content and params not allowed together')
+
+    if params and files:
+        for k,v in params.items():
+            files[k] = (None, v)
+        params = None
+        # sys.exit('cli4: params and files not allowed together')
+
+    if method != 'GET':
+        if params:
+            content = params
+            params = None
+
+    return (params, content, files)
+
 def run_command(cf, method, command, params=None, content=None, files=None):
     """run the command line"""
     # remove leading and trailing /'s
@@ -151,32 +259,27 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                         raise e
         else:
             try:
-                # dashes (vs underscores) cause issues in Python and other languages
-                if '-' in element:
+                if keyword.iskeyword(element):
+                    # a keyword is appended with an extra underscore so it can used with Python code
+                    m = getattr(m, element + '_')
+                elif '-' in element:
+                    # dashes (vs underscores) cause issues in Python and other languages
                     m = getattr(m, element.replace('-','_'))
                 else:
                     m = getattr(m, element)
                 cmd.append(element)
-            except AttributeError:
+            except AttributeError as e:
                 # the verb/element was not found
-                if len(cmd) == 0:
-                    sys.stderr.write('cli4: /%s - not found\n' % (element))
-                else:
-                    sys.stderr.write('cli4: /%s/%s - not found\n' % ('/'.join(cmd), element))
+                sys.stderr.write('cli4: /%s - not found\n' % (command))
                 raise e
 
-    if content and params:
-        sys.stderr.write('cli4: /%s - content and params not allowed together\n' % (command))
-        raise Exception
-    if content:
-        params = content
-
     results = []
     if identifier2 is None:
         identifier2 = [None]
     for i2 in identifier2:
         try:
             if method == 'GET':
+                # no content with a GET call
                 r = m.get(identifier1=identifier1,
                           identifier2=i2,
                           identifier3=identifier3,
@@ -185,22 +288,22 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                 r = m.patch(identifier1=identifier1,
                             identifier2=i2,
                             identifier3=identifier3,
-                            data=params)
+                            data=content)
             elif method == 'POST':
                 r = m.post(identifier1=identifier1,
                            identifier2=i2,
                            identifier3=identifier3,
-                           data=params, files=files)
+                           data=content, files=files)
             elif method == 'PUT':
                 r = m.put(identifier1=identifier1,
                           identifier2=i2,
                           identifier3=identifier3,
-                          data=params)
+                          data=content, files=files)
             elif method == 'DELETE':
                 r = m.delete(identifier1=identifier1,
                              identifier2=i2,
                              identifier3=identifier3,
-                             data=params)
+                             data=content)
             else:
                 pass
         except CloudFlare.exceptions.CloudFlareAPIError as e:
@@ -229,14 +332,14 @@ def write_results(results, output):
     if len(results) == 1:
         results = results[0]
 
-    if isinstance(results, str):
+    if isinstance(results, str) or isinstance(results, (bytes, bytearray)):
         # if the results are a simple string, then it should be dumped directly
         # this is only used for /zones/:id/dns_records/export presently
+        # or
+        # output is image or audio or video or something like that so we dump directly
         pass
     else:
         # anything more complex (dict, list, etc) should be dumped as JSON/YAML
-        if output is None:
-            results = None
         if output == 'json':
             try:
                 results = json.dumps(results,
@@ -249,23 +352,30 @@ def write_results(results, output):
                                      indent=4,
                                      sort_keys=True,
                                      ensure_ascii=False)
-        if output == 'yaml':
-            results = yaml.safe_dump(results)
-        if output == 'ndjson':
+        elif output == 'yaml':
+            results = my_yaml.safe_dump(results)
+        elif output == 'ndjson':
             # NDJSON support seems like a hack. There has to be a better way
             try:
-                writer = jsonlines.Writer(sys.stdout)
+                writer = my_jsonlines.Writer(sys.stdout)
                 writer.write_all(results)
                 writer.close()
             except (BrokenPipeError, IOError):
                 pass
             return
+        else:
+            # None of the above, so pass thru results except something in byte form
+            if not isinstance(results, (bytes, bytearray)):
+                results = str(results)
 
     if results:
         try:
-            sys.stdout.write(results)
-            if not results.endswith('\n'):
-                sys.stdout.write('\n')
+            if isinstance(results, (bytes, bytearray)):
+                sys.stdout.buffer.write(results)
+            else:
+                sys.stdout.write(results)
+                if not results.endswith('\n'):
+                    sys.stdout.write('\n')
         except (BrokenPipeError, IOError):
             pass
 
@@ -274,19 +384,22 @@ def do_it(args):
 
     verbose = False
     output = 'json'
+    example = False
     raw = False
     dump = False
-    dump_from_web = False
+    openapi_url = None
     binary_file = False
     profile = None
     method = 'GET'
 
     usage = ('usage: cli4 '
-             + '[-V|--version] [-h|--help] [-v|--verbose] [-q|--quiet] '
-             + '[-j|--json] [-y|--yaml] [-n|ndjson] '
+             + '[-V|--version] [-h|--help] [-v|--verbose] '
+             + '[-e|--examples] '
+             + '[-q|--quiet] '
+             + '[-j|--json] [-y|--yaml] [-n|--ndjson] [-i|--image] '
              + '[-r|--raw] '
              + '[-d|--dump] '
-             + '[-a|--api] '
+             + '[-A|--openapi url] '
              + '[-b|--binary] '
              + '[-p|--profile profile-name] '
              + '[--get|--patch|--post|--put|--delete] '
@@ -295,12 +408,15 @@ def do_it(args):
 
     try:
         opts, args = getopt.getopt(args,
-                                   'Vhvqjyrdabp:GPOUD',
+                                   'VhveqjynirdA:bp:GPOUD',
                                    [
-                                       'version',
-                                       'help', 'verbose', 'quiet', 'json', 'yaml', 'ndjson',
+                                       'version', 'help', 'verbose',
+                                       'examples',
+                                       'quiet',
+                                       'json', 'yaml', 'ndjson', 'image',
                                        'raw',
-                                       'dump', 'api',
+                                       'dump',
+                                       'openapi=',
                                        'binary',
                                        'profile=',
                                        'get', 'patch', 'post', 'put', 'delete'
@@ -316,24 +432,30 @@ def do_it(args):
             verbose = True
         elif opt in ('-q', '--quiet'):
             output = None
+        elif opt in ('-e', '--examples'):
+            example = True
         elif opt in ('-j', '--json'):
             output = 'json'
         elif opt in ('-y', '--yaml'):
-            if yaml is None:
-                sys.exit('cli4: install yaml support')
+            load_and_check_yaml()
             output = 'yaml'
         elif opt in ('-n', '--ndjson'):
-            if jsonlines is None:
+            from . import myjsonlines
+            global my_jsonlines
+            my_jsonlines = myjsonlines.myjsonlines()
+            if not my_jsonlines.available():
                 sys.exit('cli4: install jsonlines support')
             output = 'ndjson'
+        elif opt in ('-i', '--image'):
+            output = 'image'
         elif opt in ('-r', '--raw'):
             raw = True
         elif opt in ('-p', '--profile'):
             profile = arg
         elif opt in ('-d', '--dump'):
             dump = True
-        elif opt in ('-a', '--api'):
-            dump_from_web = True
+        elif opt in ('-A', '--openapi'):
+            openapi_url = arg
         elif opt in ('-b', '--binary'):
             binary_file = True
         elif opt in ('-G', '--get'):
@@ -347,121 +469,36 @@ def do_it(args):
         elif opt in ('-D', '--delete'):
             method = 'DELETE'
 
-    if dump:
-        cf = CloudFlare.CloudFlare(debug=verbose, raw=raw, profile=profile)
-        dump_commands(cf)
+    if example:
+        try:
+            examples.display()
+        except ModuleNotFoundError as e:
+            sys.exit(e)
         sys.exit(0)
 
-    if dump_from_web:
-        cf = CloudFlare.CloudFlare(debug=verbose, raw=raw, profile=profile)
-        dump_commands_from_web(cf)
-        sys.exit(0)
-
-    digits_only = re.compile('^-?[0-9]+$')
-    floats_only = re.compile('^-?[0-9.]+$')
-
-    # next grab the params. These are in the form of tag=value or =value or @filename
-    params = None
-    content = None
-    files = None
-    while len(args) > 0 and ('=' in args[0] or args[0][0] == '@'):
-        arg = args.pop(0)
-        if arg[0] == '@':
-            # a file to be uploaded - used in workers/script - only via PUT
-            filename = arg[1:]
-            if method not in ['PUT','POST']:
-                sys.exit('cli4: %s - raw file upload only with PUT or POST' % (filename))
-            try:
-                if filename == '-':
-                    if binary_file:
-                        content = sys.stdin.buffer.read()
-                    else:
-                        content = sys.stdin.read()
-                else:
-                    if binary_file:
-                        with open(filename, 'rb') as f:
-                            content = f.read()
-                    else:
-                        with open(filename, 'r') as f:
-                            content = f.read()
-            except IOError:
-                sys.exit('cli4: %s - file open failure' % (filename))
-            continue
-        tag_string, value_string = arg.split('=', 1)
-        if value_string.lower() == 'true':
-            value = True
-        elif value_string.lower() == 'false':
-            value = False
-        elif value_string == '' or value_string.lower() == 'none':
-            value = None
-        elif value_string[0] == '=' and value_string[1:] == '':
-            sys.exit('cli4: %s== - no number value passed' % (tag_string))
-        elif value_string[0] == '=' and digits_only.match(value_string[1:]):
-            value = int(value_string[1:])
-        elif value_string[0] == '=' and floats_only.match(value_string[1:]):
-            value = float(value_string[1:])
-        elif value_string[0] == '=':
-            sys.exit('cli4: %s== - invalid number value passed' % (tag_string))
-        elif value_string[0] in '[{' and value_string[-1] in '}]':
-            # a json structure - used in pagerules
-            try:
-                #value = json.loads(value) - changed to yaml code to remove unicode string issues
-                if yaml is None:
-                    sys.exit('cli4: install yaml support')
-                # cleanup string before parsing so that yaml.safe.load does not complain about whitespace
-                # >>> found character '\t' that cannot start any token <<<
-                value_string = strip_multiline(value_string)
-                try:
-                    value = yaml.safe_load(value_string)
-                except yaml.parser.ParserError as e:
-                    raise ValueError
-            except ValueError:
-                sys.exit('cli4: %s="%s" - can\'t parse json value' % (tag_string, value_string))
-        elif value_string[0] == '@':
-            # a file to be uploaded - used in dns_records/import - only via POST
-            filename = value_string[1:]
-            if method != 'POST':
-                sys.exit('cli4: %s=%s - file upload only with POST' % (tag_string, filename))
-            files = {}
-            try:
-                if filename == '-':
-                    files[tag_string] = sys.stdin
-                else:
-                    files[tag_string] = open(filename, 'rb')
-            except IOError:
-                sys.exit('cli4: %s=%s - file open failure' % (tag_string, filename))
-            # no need for param code below
-            continue
-        else:
-            value = value_string
-
-        if tag_string == '':
-            # There's no tag; it's just an unnamed list
-            if params is None:
-                params = value
-            else:
-                sys.exit('cli4: %s=%s - param error. Can\'t mix unnamed and named list' %
-                         (tag_string, value_string))
-        else:
-            if params is None:
-                params = {}
-            tag = tag_string
-            try:
-                params[tag] = value
-            except TypeError:
-                sys.exit('cli4: %s=%s - param error. Can\'t mix unnamed and named list' %
-                         (tag_string, value_string))
-
-    # what's left is the command itself
-    if len(args) < 1:
-        sys.exit(usage)
-    commands = args
-
     try:
         cf = CloudFlare.CloudFlare(debug=verbose, raw=raw, profile=profile)
     except Exception as e:
         sys.exit(e)
 
+    if dump:
+        a = dump_commands(cf)
+        sys.stdout.write(a)
+        sys.exit(0)
+
+    if openapi_url:
+        a = dump_commands_from_web(cf, openapi_url)
+        sys.stdout.write(a)
+        sys.exit(0)
+
+    # next grab the params. These are in the form of tag=value or =value or @filename
+    (params, content, files) = process_params_content_files(method, binary_file, args)
+
+    # what's left is the command itself
+    if len(args) < 1:
+        sys.exit(usage)
+    commands = args
+
     exit_with_error = False
     for command in commands:
         try:
